@@ -22,6 +22,14 @@ from .libraries.base import *
 # standard library
 from . import libraries
 
+# submodules
+from . import ast
+from . import analysis
+from . import transform
+from . import strategies
+from . import types
+from . import utils
+
 # runtime support
 from .fpc_context import FPCoreContext, NoSuchContextError
 from .interpret import (
@@ -34,16 +42,10 @@ from .interpret import (
 # compiler
 from .backend import (
     Backend,
-    CppBackend,
+    CppCompiler,
     FPCoreCompiler
 )
 
-# submodules
-from . import ast
-from . import analysis
-from . import transform
-from . import utils
-
 ###########################################################
 # typing hints
 # TODO: remove these hints
