@@ -1,6 +1,6 @@
 """Program analyses for FPy programs"""
 
-from .context_infer import ContextInfer, ContextAnalysis, ContextInferError, TypeContext
+from .context_infer import ContextInfer, ContextAnalysis, ContextInferError
 from .defs import DefAnalysis
 from .define_use import (
     DefineUse, DefineUseAnalysis,
