@@ -13,10 +13,9 @@ from ..number import Context, INTEGER, REAL
 from ..primitive import Primitive
 from ..utils import Gensym, NamedId, Unionfind
 
-from .context_types import *
+from ..types import *
 from .define_use import DefineUse, DefineUseAnalysis, Definition, DefSite
 from .type_infer import TypeInfer, TypeAnalysis
-from .types import *
 
 class ContextInferError(Exception):
     """Context inference error for FPy programs."""
@@ -24,21 +23,22 @@ class ContextInferError(Exception):
 
 @dataclass(frozen=True)
 class ContextAnalysis:
-    func_ty: FunctionTypeContext
-    by_def: dict[Definition, TypeContext]
-    by_expr: dict[Expr, TypeContext]
+    fn_type: FunctionType
+    by_def: dict[Definition, Type]
+    by_expr: dict[Expr, Type]
+    at_expr: dict[Expr, ContextParam]
 
     @property
     def body_ctx(self):
-        return self.func_ty.ctx
+        return self.fn_type.ctx
 
     @property
     def arg_types(self):
-        return self.func_ty.args
+        return self.fn_type.arg_types
 
     @property
     def return_type(self):
-        return self.func_ty.ret
+        return self.fn_type.return_type
 
 
 class ContextTypeInferInstance(Visitor):
@@ -54,9 +54,10 @@ class ContextTypeInferInstance(Visitor):
     type_info: TypeAnalysis
     unsafe_cast_int: bool
 
-    by_def: dict[Definition, TypeContext]
-    by_expr: dict[Expr, TypeContext]
-    ret_ty: TypeContext | None
+    by_def: dict[Definition, Type]
+    by_expr: dict[Expr, Type]
+    at_expr: dict[Expr, ContextParam]
+    ret_ty: Type | None
     rvars: Unionfind[ContextParam]
     gensym: Gensym
 
@@ -73,6 +74,7 @@ class ContextTypeInferInstance(Visitor):
         self.unsafe_cast_int = unsafe_cast_int
         self.by_def = {}
         self.by_expr = {}
+        self.at_expr = {}
         self.ret_ty = None
         self.rvars = Unionfind()
         self.gensym = Gensym()
@@ -83,15 +85,15 @@ class ContextTypeInferInstance(Visitor):
     def _from_scalar(self, ty: Type, ctx: ContextParam):
         match ty:
             case BoolType():
-                return BoolTypeContext()
+                return BoolType()
             case RealType():
-                return RealTypeContext(ctx)
+                return RealType(ctx)
             case ContextType():
-                return ContextTypeContext()
+                return ContextType()
             case _:
                 raise RuntimeError(f'unreachable: {ty}')
 
-    def _set_context(self, site: Definition, ty: TypeContext):
+    def _set_context(self, site: Definition, ty: Type):
         self.by_def[site] = ty
 
     def _fresh_context_var(self) -> NamedId:
@@ -102,38 +104,39 @@ class ContextTypeInferInstance(Visitor):
     def _resolve_context(self, ctx: ContextParam) -> ContextParam:
         return self.rvars.get(ctx, ctx)
 
-    def _resolve(self, ty: TypeContext) -> TypeContext:
+    def _resolve(self, ty: Type) -> Type:
         match ty:
-            case BoolTypeContext() | ContextTypeContext() | VarTypeContext():
+            case BoolType() | ContextType() | VarType():
                 return ty
-            case RealTypeContext():
+            case RealType():
+                assert ty.ctx is not None
                 ctx = self._resolve_context(ty.ctx)
-                return RealTypeContext(ctx)
-            case TupleTypeContext():
+                return RealType(ctx)
+            case TupleType():
                 elts = (self._resolve(elt) for elt in ty.elts)
-                return TupleTypeContext(*elts)
-            case ListTypeContext():
+                return TupleType(*elts)
+            case ListType():
                 elt = self._resolve(ty.elt)
-                return ListTypeContext(elt)
+                return ListType(elt)
             case _:
                 raise RuntimeError(f'unreachable: {ty}')
 
-    def _instantiate(self, ty: TypeContext) -> TypeContext:
+    def _instantiate(self, ty: Type) -> Type:
         subst: dict[NamedId, ContextParam] = {}
-        for fv in sorted(ty.free_vars()):
+        for fv in sorted(ty.free_context_vars()):
             subst[fv] = self._fresh_context_var()
-        return ty.subst(subst)
+        return ty.subst_context(subst)
 
-    def _generalize(self, ty: TypeContext) -> tuple[TypeContext, dict[NamedId, ContextParam]]:
+    def _generalize(self, ty: Type) -> tuple[Type, dict[NamedId, ContextParam]]:
         subst: dict[NamedId, ContextParam] = {}
-        for i, fv in enumerate(sorted(ty.free_vars())):
+        for i, fv in enumerate(sorted(ty.free_context_vars())):
             t = self.rvars.find(fv)
             match t: 
                 case NamedId():
                     subst[fv] = NamedId(f'r{i + 1}')
                 case _:
                     subst[fv] = t
-        ty = ty.subst(subst)
+        ty = ty.subst_context(subst)
         return ty, subst
 
     def _unify_contexts(self, a: ContextParam, b: ContextParam) -> ContextParam:
@@ -151,47 +154,79 @@ class ContextTypeInferInstance(Visitor):
             case _:
                 raise RuntimeError(f'unreachable case: {a}, {b}')
 
-    def _unify(self, a_ty: TypeContext, b_ty: TypeContext) -> TypeContext:
+    def _unify(self, a_ty: Type, b_ty: Type) -> Type:
         match a_ty, b_ty:
-            case VarTypeContext(), VarTypeContext():
+            case VarType(), VarType():
                 if a_ty != b_ty:
                     raise ContextInferError(f'incompatible types: {a_ty} != {b_ty}')
                 return a_ty
-            case (BoolTypeContext(), BoolTypeContext()) | (ContextTypeContext(), ContextTypeContext()):
+            case (BoolType(), BoolType()) | (ContextType(), ContextType()):
                 return a_ty
-            case RealTypeContext(), RealTypeContext():
+            case RealType(), RealType():
+                assert a_ty.ctx is not None and b_ty.ctx is not None
                 ctx = self._unify_contexts(a_ty.ctx, b_ty.ctx)
-                return RealTypeContext(ctx)
-            case TupleTypeContext(), TupleTypeContext():
+                return RealType(ctx)
+            case TupleType(), TupleType():
                 assert len(a_ty.elts) == len(b_ty.elts)
                 elts = [self._unify(a_elt, b_elt) for a_elt, b_elt in zip(a_ty.elts, b_ty.elts)]
-                return TupleTypeContext(*elts)
-            case ListTypeContext(), ListTypeContext():
+                return TupleType(*elts)
+            case ListType(), ListType():
                 elt = self._unify(a_ty.elt, b_ty.elt)
-                return ListTypeContext(elt)
+                return ListType(elt)
             case _:
                 raise RuntimeError(f'unreachable: {a_ty}, {b_ty}')
 
-    def _cvt_arg_type(self, ty: Type):
+    def _cvt_type(self, ty: Type) -> Type:
         match ty:
             case VarType():
-                return VarTypeContext(ty.name)
+                return VarType(ty.name)
             case BoolType():
-                return BoolTypeContext()
+                return BoolType()
             case RealType():
-                return RealTypeContext(self._fresh_context_var())
+                return RealType(self._fresh_context_var())
             case ContextType():
-                return ContextTypeContext()
+                return ContextType()
             case TupleType():
-                elts = [self._cvt_arg_type(elt) for elt in ty.elts]
-                return TupleTypeContext(*elts)
+                elts = [self._cvt_type(elt) for elt in ty.elts]
+                return TupleType(*elts)
             case ListType():
-                elt = self._cvt_arg_type(ty.elt)
-                return ListTypeContext(elt)
+                elt = self._cvt_type(ty.elt)
+                return ListType(elt)
             case _:
                 raise RuntimeError(f'unreachable: {ty}')
 
-    def _visit_binding(self, site: DefSite, target: Id | TupleBinding, ty: TypeContext):
+    def _cvt_arg_type(self, ty: Type, ann: TypeAnn):
+        match ty:
+            case VarType():
+                return VarType(ty.name)
+            case BoolType():
+                return BoolType()
+            case RealType():
+                if isinstance(ann, RealTypeAnn) and ann.ctx is not None:
+                    return RealType(ann.ctx)
+                else:
+                    return RealType(self._fresh_context_var())
+            case ContextType():
+                return ContextType()
+            case TupleType():
+                if isinstance(ann, TupleTypeAnn):
+                    assert len(ann.elts) == len(ty.elts)
+                    elts = [self._cvt_arg_type(elt, ann) for elt, ann in zip(ty.elts, ann.elts)]
+                    return TupleType(*elts)
+                else:
+                    elts = [self._cvt_arg_type(elt, AnyTypeAnn(None)) for elt in ty.elts]
+                    return TupleType(*elts)
+            case ListType():
+                if isinstance(ann, ListTypeAnn):
+                    elt = self._cvt_arg_type(ty.elt, ann.elt)
+                    return ListType(elt)
+                else:
+                    elt = self._cvt_arg_type(ty.elt, AnyTypeAnn(None))
+                    return ListType(elt)
+            case _:
+                raise RuntimeError(f'unreachable: {ty}')
+
+    def _visit_binding(self, site: DefSite, target: Id | TupleBinding, ty: Type):
         match target:
             case NamedId():
                 d = self.def_use.find_def_from_site(target, site)
@@ -199,7 +234,7 @@ class ContextTypeInferInstance(Visitor):
             case UnderscoreId():
                 pass
             case TupleBinding():
-                assert isinstance(ty, TupleTypeContext) and len(ty.elts) == len(target.elts)
+                assert isinstance(ty, TupleType) and len(ty.elts) == len(target.elts)
                 for elt, elt_ctx in zip(target.elts, ty.elts):
                     self._visit_binding(site, elt, elt_ctx)
             case _:
@@ -215,65 +250,65 @@ class ContextTypeInferInstance(Visitor):
     def _visit_bool(self, e: BoolVal, ctx: ContextParam):
         # C, Γ |- e : bool
         ty = self._from_scalar(self._lookup_ty(e), ctx)
-        assert isinstance(ty, BoolTypeContext) # type checking should have concluded this
+        assert isinstance(ty, BoolType) # type checking should have concluded this
         return ty
 
     def _visit_foreign(self, e: ForeignVal, ctx: ContextParam):
-        return self._cvt_arg_type(self._lookup_ty(e))
+        return self._cvt_type(self._lookup_ty(e))
 
     def _visit_decnum(self, e: Decnum, ctx: ContextParam):
         if self.unsafe_cast_int and e.is_integer():
             # unsafely cast to integer
             # C, Γ |- e : real INTEGER
-            return RealTypeContext(INTEGER)
+            return RealType(INTEGER)
         else:
             # C, Γ |- e : real REAL
             ty = self._from_scalar(self._lookup_ty(e), REAL)
-            assert isinstance(ty, RealTypeContext) # type checking should have concluded this
+            assert isinstance(ty, RealType) # type checking should have concluded this
             return ty
 
     def _visit_hexnum(self, e: Hexnum, ctx: ContextParam):
         if self.unsafe_cast_int and e.is_integer():
             # unsafely cast to integer
             # C, Γ |- e : real INTEGER
-            return RealTypeContext(INTEGER)
+            return RealType(INTEGER)
         else:
             # C, Γ |- e : real REAL
             ty = self._from_scalar(self._lookup_ty(e), REAL)
-            assert isinstance(ty, RealTypeContext) # type checking should have concluded this
+            assert isinstance(ty, RealType) # type checking should have concluded this
             return ty
 
     def _visit_integer(self, e: Integer, ctx: ContextParam):
         if self.unsafe_cast_int and e.is_integer():
             # unsafely cast to integer
             # C, Γ |- e : real INTEGER
-            return RealTypeContext(INTEGER)
+            return RealType(INTEGER)
         else:
             # C, Γ |- e : real REAL
             ty = self._from_scalar(self._lookup_ty(e), REAL)
-            assert isinstance(ty, RealTypeContext) # type checking should have concluded this
+            assert isinstance(ty, RealType) # type checking should have concluded this
             return ty
 
     def _visit_rational(self, e: Rational, ctx: ContextParam):
         if self.unsafe_cast_int and e.is_integer():
             # unsafely cast to integer
             # C, Γ |- e : real INTEGER
-            return RealTypeContext(INTEGER)
+            return RealType(INTEGER)
         else:
             # C, Γ |- e : real REAL
             ty = self._from_scalar(self._lookup_ty(e), REAL)
-            assert isinstance(ty, RealTypeContext) # type checking should have concluded this
+            assert isinstance(ty, RealType) # type checking should have concluded this
             return ty
 
     def _visit_digits(self, e: Digits, ctx: ContextParam):
         if self.unsafe_cast_int and e.is_integer():
             # unsafely cast to integer
             # C, Γ |- e : real INTEGER
-            return RealTypeContext(INTEGER)
+            return RealType(INTEGER)
         else:
             # C, Γ |- e : real REAL
             ty = self._from_scalar(self._lookup_ty(e), REAL)
-            assert isinstance(ty, RealTypeContext) # type checking should have concluded this
+            assert isinstance(ty, RealType) # type checking should have concluded this
             return ty
 
     def _visit_nullaryop(self, e: NullaryOp, ctx: ContextParam):
@@ -288,27 +323,27 @@ class ContextTypeInferInstance(Visitor):
             case Len() | Dim():
                 # length / dimension
                 # C, Γ |- len e : real INTEGER
-                return RealTypeContext(INTEGER)
+                return RealType(INTEGER)
             case Sum():
                 # sum operator
                 # C, Γ |- sum e : real C
-                return RealTypeContext(ctx)
-            case Range():
+                return RealType(ctx)
+            case Range1():
                 # range operator
                 # C, Γ |- range e : list (real INTEGER)
-                return ListTypeContext(RealTypeContext(INTEGER))
+                return ListType(RealType(INTEGER))
             case Empty():
                 # empty operator
                 # C, Γ |- empty e : list T
-                return self._cvt_arg_type(self._lookup_ty(e))
+                return self._cvt_type(self._lookup_ty(e))
             case Enumerate():
                 # enumerate operator
                 #          C, Γ |- e : list T
                 # -----------------------------------------
                 #  C, Γ |- enumerate e : list [real INTEGER] x T
-                assert isinstance(arg_ty, ListTypeContext)
-                elt_ty = TupleTypeContext(RealTypeContext(INTEGER), arg_ty.elt)
-                return ListTypeContext(elt_ty)
+                assert isinstance(arg_ty, ListType)
+                elt_ty = TupleType(RealType(INTEGER), arg_ty.elt)
+                return ListType(elt_ty)
             case _:
                 #   Γ |- real : T         Γ |- bool : T
                 # ----------------      ------------------
@@ -322,7 +357,11 @@ class ContextTypeInferInstance(Visitor):
             case Size():
                 # size operator
                 # C, Γ |- size e : real INTEGER
-                return RealTypeContext(INTEGER)
+                return RealType(INTEGER)
+            case Range2():
+                # range operator
+                # C, Γ |- range e1 e2 : list (real INTEGER)
+                return ListType(RealType(INTEGER))
             case _:
                 #   Γ |- real : T         Γ |- bool : T
                 # ----------------      ------------------
@@ -330,13 +369,19 @@ class ContextTypeInferInstance(Visitor):
                 return self._from_scalar(self._lookup_ty(e), ctx)
 
     def _visit_ternaryop(self, e: TernaryOp, ctx: ContextParam):
-        #   Γ |- real : T         Γ |- bool : T
-        # ----------------      ------------------
-        #  C, Γ |- e : real C    C, Γ |- e : bool
         self._visit_expr(e.first, ctx)
         self._visit_expr(e.second, ctx)
         self._visit_expr(e.third, ctx)
-        return self._from_scalar(self._lookup_ty(e), ctx)
+        match e:
+            case Range3():
+                # range operator
+                # C, Γ |- range e1 e2 e3 : list (real INTEGER)
+                return ListType(RealType(INTEGER))
+            case _:
+                #   Γ |- real : T         Γ |- bool : T
+                # ----------------      ------------------
+                #  C, Γ |- e : real C    C, Γ |- e : bool
+                return self._from_scalar(self._lookup_ty(e), ctx)
 
     def _visit_naryop(self, e: NaryOp, ctx: ContextParam):
         arg_tys = [self._visit_expr(arg, ctx) for arg in e.args]
@@ -353,7 +398,7 @@ class ContextTypeInferInstance(Visitor):
             case And() | Or():
                 # and / or operator
                 # C, Γ |- e : bool
-                return BoolTypeContext()
+                return BoolType()
             case Zip():
                 # zip operator
                 #  C, Γ |- e_1 : list T_1 ... C, Γ |- e_n : list T_n
@@ -361,9 +406,9 @@ class ContextTypeInferInstance(Visitor):
                 #        C, Γ |- e : list [T_1 x ... x T_n]
                 elt_tys = []
                 for arg_ty in arg_tys:
-                    assert isinstance(arg_ty, ListTypeContext)
+                    assert isinstance(arg_ty, ListType)
                     elt_tys.append(arg_ty.elt)
-                return ListTypeContext(TupleTypeContext(*elt_tys))
+                return ListType(TupleType(*elt_tys))
             case _:
                 raise ValueError(f'unknown n-ary operator: {type(e)}')
 
@@ -371,7 +416,7 @@ class ContextTypeInferInstance(Visitor):
         # C, Γ |- e : bool
         for arg in e.args:
             self._visit_expr(arg, ctx)
-        return BoolTypeContext()
+        return BoolType()
 
     def _visit_call(self, e: Call, ctx: ContextParam):
         # get around circular imports
@@ -380,49 +425,54 @@ class ContextTypeInferInstance(Visitor):
         match e.fn:
             case None:
                 # calling None => can't conclude anything
-                ty = self._cvt_arg_type(self._lookup_ty(e))
+                ty = self._cvt_type(self._lookup_ty(e))
                 return ty
             case Primitive():
                 # calling a primitive => can't conclude anything
                 fn_ty = ContextInfer.infer_primitive(e.fn)
                 # instantiate the function context
-                fn_ty = cast(FunctionTypeContext, self._instantiate(fn_ty))
+                fn_ty = cast(FunctionType, self._instantiate(fn_ty))
                 # merge caller context
+                assert fn_ty.ctx is not None
                 self._unify_contexts(ctx, fn_ty.ctx)
                 # merge arguments
-                if len(fn_ty.args) != len(e.args):
+                if len(fn_ty.arg_types) != len(e.args):
                     raise ContextInferError(
-                        f'primitive {e.fn} expects {len(fn_ty.args)} arguments, '
+                        f'primitive {e.fn} expects {len(fn_ty.arg_types)} arguments, '
                         f'got {len(e.args)}'
                     )
-                for arg, expect_ty in zip(e.args, fn_ty.args):
+                for arg, expect_ty in zip(e.args, fn_ty.arg_types):
                     ty = self._visit_expr(arg, ctx)
                     self._unify(ty, expect_ty)
 
-                return fn_ty.ret
+                return fn_ty.return_type
             case Function():
                 # calling a function
                 # TODO: guard against recursion
-                from ..transform import ConstFold
+                from ..transform import ConstFold, Monomorphize
 
+                # get argument types
+                arg_types = [self.type_info.by_expr[arg] for arg in e.args]
+
+                # TODO: is there a better way to do this? bad complexity
+                # apply constant folding and monomorphization
                 ast = ConstFold.apply(e.fn.ast, enable_op=False)
+                ast = Monomorphize.apply_by_arg(ast, ctx if isinstance(ctx, Context) else None, arg_types)
                 fn_info = ContextInfer.infer(ast)
-                if len(fn_info.arg_types) != len(e.args):
-                    raise ContextInferError(
-                        f'function {e.fn} expects {len(fn_info.arg_types)} arguments, '
-                        f'got {len(e.args)}'
-                    )
 
                 # instantiate the function context
-                fn_ctx = cast(FunctionTypeContext, self._instantiate(fn_info.func_ty))
+                fn_type = cast(FunctionType, self._instantiate(fn_info.fn_type))
+
                 # merge caller context
-                self._unify_contexts(ctx, fn_ctx.ctx)
+                assert fn_type.ctx is not None
+                self._unify_contexts(ctx, fn_type.ctx)
+
                 # merge arguments
-                for arg, expect_ty in zip(e.args, fn_ctx.args):
+                for arg, expect_ty in zip(e.args, fn_type.arg_types):
                     ty = self._visit_expr(arg, ctx)
                     self._unify(ty, expect_ty)
 
-                return fn_ctx.ret
+                return fn_type.return_type
             case type() if issubclass(e.fn, Context):
                 # calling context constructor
                 # TODO: can infer if the arguments are statically known
@@ -435,18 +485,18 @@ class ContextTypeInferInstance(Visitor):
         # -----------------------------------------
         #        C, Γ |- e : T_1 x ... x T_n
         arg_tys = [self._visit_expr(arg, ctx) for arg in e.elts]
-        return TupleTypeContext(*arg_tys)
+        return TupleType(*arg_tys)
 
     def _visit_list_expr(self, e: ListExpr, ctx: ContextParam):
         #  C, Γ |- e_1 : T ... C, Γ |- e_n : T
         # -------------------------------------
         #         C, Γ |- e : list T
         if len(e.elts) == 0:
-            return self._cvt_arg_type(self._lookup_ty(e))
+            return self._cvt_type(self._lookup_ty(e))
         else:
             # type checking ensures the base type is the same
             elts = [self._visit_expr(arg, ctx) for arg in e.elts]
-            return ListTypeContext(elts[0])
+            return ListType(elts[0])
 
     def _visit_list_comp(self, e: ListComp, ctx: ContextParam):
         #       C, Γ |- elt: T
@@ -454,10 +504,10 @@ class ContextTypeInferInstance(Visitor):
         #  C, Γ |- for ... : list T
         for target, iterable in zip(e.targets, e.iterables):
             iterable_ty = self._visit_expr(iterable, ctx)
-            assert isinstance(iterable_ty, ListTypeContext)
+            assert isinstance(iterable_ty, ListType)
             self._visit_binding(e, target, iterable_ty.elt)
         ty = self._visit_expr(e.elt, ctx)
-        return ListTypeContext(ty)
+        return ListType(ty)
 
     def _visit_list_ref(self, e: ListRef, ctx: ContextParam):
         #      C, Γ |- e: list T       C, Γ |- i : real C
@@ -465,7 +515,7 @@ class ContextTypeInferInstance(Visitor):
         #            C, Γ |- e[i] : T
         ty = self._visit_expr(e.value, ctx)
         self._visit_expr(e.index, ctx)
-        assert isinstance(ty, ListTypeContext)
+        assert isinstance(ty, ListType)
         return ty.elt
 
     def _visit_list_slice(self, e: ListSlice, ctx: ContextParam):
@@ -559,7 +609,7 @@ class ContextTypeInferInstance(Visitor):
 
     def _visit_for(self, stmt: ForStmt, ctx: ContextParam):
         iter_ty = self._visit_expr(stmt.iterable, ctx)
-        assert isinstance(iter_ty, ListTypeContext)
+        assert isinstance(iter_ty, ListType)
         self._visit_binding(stmt, stmt.target, iter_ty.elt)
 
         # add types to phi variables
@@ -623,9 +673,9 @@ class ContextTypeInferInstance(Visitor):
                 body_ctx = func.ctx
 
         # generate context variables for each argument
-        arg_types: list[TypeContext] = []
+        arg_types: list[Type] = []
         for arg, ty in zip(func.args, self.type_info.arg_types):
-            arg_ty = self._cvt_arg_type(ty)
+            arg_ty = self._cvt_arg_type(ty, arg.type)
             arg_types.append(arg_ty)
             if isinstance(arg.name, NamedId):
                 d = self.def_use.find_def_from_site(arg.name, arg)
@@ -634,21 +684,22 @@ class ContextTypeInferInstance(Visitor):
         # generate context variables for each free variables
         for v in func.free_vars:
             d = self.def_use.find_def_from_site(v, func)
-            ty = self._cvt_arg_type(self.type_info.by_def[d])
+            ty = self._cvt_type(self.type_info.by_def[d])
             self._set_context(d, ty)
 
         # visit body
         self._visit_block(func.body, body_ctx)
-        assert isinstance(self.ret_ty, TypeContext) # function has no return statement
+        assert isinstance(self.ret_ty, Type) # function has no return statement
 
         # generalize the function context
         arg_types = [self._resolve(ty) for ty in arg_types]
         ret_ty = self._resolve(self.ret_ty)
-        return FunctionTypeContext(body_ctx, arg_types, ret_ty)
+        return FunctionType(body_ctx, arg_types, ret_ty)
 
-    def _visit_expr(self, expr: Expr, ctx: ContextParam) -> TypeContext:
+    def _visit_expr(self, expr: Expr, ctx: ContextParam) -> Type:
         ty = super()._visit_expr(expr, ctx)
         self.by_expr[expr] = ty
+        self.at_expr[expr] = ctx
         return ty
 
     def infer(self):
@@ -657,7 +708,7 @@ class ContextTypeInferInstance(Visitor):
 
         # generalize the output context
         fn_ctx, subst = self._generalize(ctx)
-        fn_ctx = cast(FunctionTypeContext, fn_ctx)
+        fn_ctx = cast(FunctionType, fn_ctx)
 
         # rename unbound context variables
         for t in self.rvars:
@@ -666,14 +717,19 @@ class ContextTypeInferInstance(Visitor):
 
         # resolve definition/expr contexts
         by_defs = {
-            d: self._resolve(ctx).subst(subst)
+            d: self._resolve(ctx).subst_context(subst)
             for d, ctx in self.by_def.items()
         }
         by_expr = {
-            e: self._resolve(ctx).subst(subst)
+            e: self._resolve(ctx).subst_context(subst)
             for e, ctx in self.by_expr.items()
         }
-        return ContextAnalysis(fn_ctx, by_defs, by_expr)
+        at_expr = {
+            e: self._resolve_context(ctx)
+            for e, ctx in self.at_expr.items()
+        }
+
+        return ContextAnalysis(fn_ctx, by_defs, by_expr, at_expr)
 
 
 class _ContextInferPrimitive:
@@ -696,83 +752,83 @@ class _ContextInferPrimitive:
     def _fresh_context_var(self) -> NamedId:
         return self.gensym.fresh('r')
 
-    def _cvt_arg_type(self, ty: Type, ctx: str | tuple | None) -> TypeContext:
+    def _cvt_arg_type(self, ty: Type, ctx: str | tuple | None) -> Type:
         match ty:
             case VarType():
-                return VarTypeContext(ty.name)
+                return VarType(ty.name)
             case BoolType():
-                return BoolTypeContext()
+                return BoolType()
             case RealType():
                 if ctx is None:
-                    return RealTypeContext(self._fresh_context_var())
+                    return RealType(self._fresh_context_var())
                 else:
                     if not isinstance(ctx, str):
                         raise ValueError(f"expected context variable for argument of type {ty}, got {ctx}")
                     if ctx not in self.subst:
                         self.subst[ctx] = self._fresh_context_var()
-                    return RealTypeContext(self.subst[ctx])
+                    return RealType(self.subst[ctx])
             case ContextType():
-                return ContextTypeContext()
+                return ContextType()
             case TupleType():
                 if ctx is None:
                     elts = [self._cvt_arg_type(t, None) for t in ty.elts]
-                    return TupleTypeContext(*elts)
+                    return TupleType(*elts)
                 else:
                     if not isinstance(ctx, tuple):
                         raise ValueError(f"expected tuple context for argument of type {ty}, got {ctx}")
                     if len(ty.elts) != len(ctx):
                         raise ValueError(f"tuple context length mismatch: expected {len(ty.elts)}, got {len(ctx)}")
                     elts = [self._cvt_arg_type(t, c) for t, c in zip(ty.elts, ctx)]
-                    return TupleTypeContext(*elts)
+                    return TupleType(*elts)
             case ListType():
                 if ctx is None:
                     elt = self._cvt_arg_type(ty.elt, None)
-                    return ListTypeContext(elt)
+                    return ListType(elt)
                 else:
-                    return ListTypeContext(self._cvt_arg_type(ty.elt, ctx))
+                    return ListType(self._cvt_arg_type(ty.elt, ctx))
             case _:
                 raise RuntimeError(f'unknown type: {ty}')
 
-    def _cvt_ret_type(self, ty: Type, ctx: Context | str | tuple | None) -> TypeContext:
+    def _cvt_ret_type(self, ty: Type, ctx: Context | str | tuple | None) -> Type:
         match ty:
             case VarType():
-                return VarTypeContext(ty.name)
+                return VarType(ty.name)
             case BoolType():
-                return BoolTypeContext()
+                return BoolType()
             case RealType():
                 if ctx is None:
-                    return RealTypeContext(self._fresh_context_var())
+                    return RealType(self._fresh_context_var())
                 elif isinstance(ctx, Context):
-                    return RealTypeContext(ctx)
+                    return RealType(ctx)
                 else:
                     if not isinstance(ctx, str):
                         raise ValueError(f"expected context variable for return of type {ty}, got {ctx}")
                     if ctx not in self.subst:
                         raise ValueError(f"unbound context variable '{ctx}' in return type")
-                    return RealTypeContext(self.subst[ctx])
+                    return RealType(self.subst[ctx])
             case ContextType():
-                return ContextTypeContext()
+                return ContextType()
             case TupleType():
                 if ctx is None:
                     elts = [self._cvt_ret_type(t, None) for t in ty.elts]
-                    return TupleTypeContext(*elts)
+                    return TupleType(*elts)
                 else:
                     if not isinstance(ctx, tuple):
                         raise ValueError(f"expected tuple context for return of type {ty}, got {ctx}")
                     if len(ty.elts) != len(ctx):
                         raise ValueError(f"tuple context length mismatch: expected {len(ty.elts)}, got {len(ctx)}")
                     elts = [self._cvt_ret_type(t, c) for t, c in zip(ty.elts, ctx)]
-                    return TupleTypeContext(*elts)
+                    return TupleType(*elts)
             case ListType():
                 if ctx is None:
                     elt = self._cvt_ret_type(ty.elt, None)
-                    return ListTypeContext(elt)
+                    return ListType(elt)
                 else:
-                    return ListTypeContext(self._cvt_ret_type(ty.elt, ctx)) 
+                    return ListType(self._cvt_ret_type(ty.elt, ctx)) 
             case _:
                 raise RuntimeError(f'unknown type: {ty}')
 
-    def infer(self) -> FunctionTypeContext:
+    def infer(self) -> FunctionType:
         # perform standard type inference
         fn_ty = TypeInfer.infer_primitive(self.prim)
 
@@ -792,7 +848,7 @@ class _ContextInferPrimitive:
         # interpret return context
         ret_type = self._cvt_ret_type(fn_ty.return_type, self.prim.ret_ctx)
 
-        return FunctionTypeContext(ctx, arg_types, ret_type)
+        return FunctionType(ctx, arg_types, ret_type)
 
 
 ###########################################################
@@ -839,7 +895,7 @@ class ContextInfer:
         return inst.infer()
 
     @staticmethod
-    def infer_primitive(prim: Primitive) -> FunctionTypeContext:
+    def infer_primitive(prim: Primitive) -> FunctionType:
         """
         Infers the context of a primitive.
         """
