@@ -36,10 +36,9 @@ __all__ = [
 
     # Type annotations
     'AnyTypeAnn',
-    'ScalarTypeAnn',
     'RealTypeAnn',
     'BoolTypeAnn',
-    'TensorTypeAnn',
+    'ContextTypeAnn',
     'TupleTypeAnn',
     'ListTypeAnn',
     'SizedTensorTypeAnn',
@@ -103,6 +102,7 @@ __all__ = [
     # Other arithmetic
     'Max',
     'Min',
+    'Mod',
     'Fmod',
     'Remainder',
     'Cbrt',
@@ -168,7 +168,9 @@ __all__ = [
     # Tensor operators
     'Len',
     'Size',
-    'Range',
+    'Range1',
+    'Range2',
+    'Range3',
     'Empty',
     'Dim',
     'Zip',
@@ -242,6 +244,17 @@ class TypeAnn(Ast):
     def __init__(self, loc: Location | None):
         super().__init__(loc)
 
+    @abstractmethod
+    def is_equiv(self, other) -> bool:
+        """
+        Check if this type annotation is equivalent to another type annotation.
+
+        This is essentially a recursive equality check.
+        The dunder method `__eq__` is used to check if two expressions
+        represent exactly the same tree, e.g., `id(self) == id(other)`.
+        """
+        ...
+
 class AnyTypeAnn(TypeAnn):
     """FPy AST: any type annotation"""
     __slots__ = ()
@@ -249,31 +262,21 @@ class AnyTypeAnn(TypeAnn):
     def __init__(self, loc: Location | None):
         super().__init__(loc)
 
-    def __eq__(self, other: object) -> bool:
+    def is_equiv(self, other):
         return isinstance(other, AnyTypeAnn)
 
-    def __hash__(self) -> int:
-        return hash(())
-
-class ScalarTypeAnn(TypeAnn):
-    """FPy AST: scalar type annotation"""
-    __slots__ = ()
-
-    def __init__(self, loc: Location | None):
-        super().__init__(loc)
-
 class RealTypeAnn(TypeAnn):
     """FPy AST: real type annotation"""
-    __slots__ = ()
+    __slots__ = ('ctx',)
 
-    def __init__(self, loc: Location | None):
+    ctx: Context | None
+
+    def __init__(self, ctx: Context | None, loc: Location | None):
         super().__init__(loc)
+        self.ctx = ctx
 
-    def __eq__(self, other: object) -> bool:
-        return isinstance(other, RealTypeAnn)
-
-    def __hash__(self) -> int:
-        return hash(())
+    def is_equiv(self, other):
+        return isinstance(other, RealTypeAnn) and self.ctx == other.ctx
 
 class BoolTypeAnn(TypeAnn):
     """FPy AST: boolean type annotation"""
@@ -282,20 +285,21 @@ class BoolTypeAnn(TypeAnn):
     def __init__(self, loc: Location | None):
         super().__init__(loc)
 
-    def __eq__(self, other: object) -> bool:
+    def is_equiv(self, other):
         return isinstance(other, BoolTypeAnn)
 
-    def __hash__(self) -> int:
-        return hash(())
-
-class TensorTypeAnn(TypeAnn):
-    """FPy AST: tensor type annotation"""
+class ContextTypeAnn(TypeAnn):
+    """FPy AST: context type annotation"""
     __slots__ = ()
 
     def __init__(self, loc: Location | None):
         super().__init__(loc)
 
-class TupleTypeAnn(TensorTypeAnn):
+    def is_equiv(self, other):
+        return isinstance(other, ContextTypeAnn)
+
+
+class TupleTypeAnn(TypeAnn):
     """FPy AST: native tuple type annotation"""
     __slots__ = ('elts',)
     elts: tuple[TypeAnn, ...]
@@ -304,15 +308,14 @@ class TupleTypeAnn(TensorTypeAnn):
         super().__init__(loc)
         self.elts = tuple(elts)
 
-    def __eq__(self, other: object) -> bool:
-        if not isinstance(other, TupleTypeAnn):
-            return False
-        return self.elts == other.elts
+    def is_equiv(self, other) -> bool:
+        return (
+            isinstance(other, TupleTypeAnn)
+            and len(self.elts) == len(other.elts)
+            and all(a.is_equiv(b) for a, b in zip(self.elts, other.elts))
+        )
 
-    def __hash__(self) -> int:
-        return hash(self.elts)
-
-class ListTypeAnn(TensorTypeAnn):
+class ListTypeAnn(TypeAnn):
     """FPy AST: native list type annotation"""
     __slots__ = ('elt',)
     elt: TypeAnn
@@ -321,15 +324,10 @@ class ListTypeAnn(TensorTypeAnn):
         super().__init__(loc)
         self.elt = elt
 
-    def __eq__(self, other: object) -> bool:
-        if not isinstance(other, ListTypeAnn):
-            return False
-        return self.elt == other.elt
+    def is_equiv(self, other) -> bool:
+        return isinstance(other, ListTypeAnn) and self.elt.is_equiv(other.elt)
 
-    def __hash__(self) -> int:
-        return hash(self.elt)
-
-class SizedTensorTypeAnn(TensorTypeAnn):
+class SizedTensorTypeAnn(TypeAnn):
     """FPy AST: sized, homogenous tensor type annotation"""
     __slots__ = ('dims', 'elt')
     dims: tuple[int | NamedId, ...]
@@ -340,13 +338,13 @@ class SizedTensorTypeAnn(TensorTypeAnn):
         self.dims = tuple(dims)
         self.elt = elt
 
-    def __eq__(self, other: object) -> bool:
-        if not isinstance(other, SizedTensorTypeAnn):
-            return False
-        return self.dims == other.dims and self.elt == other.elt
+    def is_equiv(self, other) -> bool:
+        return (
+            isinstance(other, SizedTensorTypeAnn)
+            and self.dims == other.dims
+            and self.elt.is_equiv(other.elt)
+        )
 
-    def __hash__(self) -> int:
-        return hash((self.dims, self.elt))
 
 class Expr(Ast):
     """FPy AST: expression"""
@@ -857,12 +855,40 @@ class Min(NamedNaryOp):
     """FPy node: `min(x, y, ...)`"""
     __slots__ = ()
 
+class Mod(BinaryOp):
+    """
+    FPy node: `a % b`
+
+    Represents Python's modulus operator, defined as:
+
+        a = (a // b) * b + (a % b)
+
+    The result has the same sign as the divisor `b`.
+    """
+    __slots__ = ()
+
 class Fmod(NamedBinaryOp):
-    """FPy node: modulus"""
+    """
+    FPy node: `fmod(x, y)`
+
+    Represents the C standard library's `fmod` function, defined as:
+
+        fmod(x, y) = x - trunc(x / y) * y
+
+    The result has the same sign as the dividend `x`.
+    """
     __slots__ = ()
 
 class Remainder(NamedBinaryOp):
-    """FPy node: remainder"""
+    """
+    FPy node: `remainder(x, y)`
+
+    Represents the IEEE 754 `remainder` function, defined as:
+
+        remainder(x, y) = x - round(x / y) * y
+
+    The result is the value closest to zero.
+    """
     __slots__ = ()
 
 class Cbrt(NamedUnaryOp):
@@ -1063,8 +1089,16 @@ class Size(NamedBinaryOp):
     """FPy node: size operator"""
     __slots__ = ()
 
-class Range(NamedUnaryOp):
-    """FPy node: range constructor"""
+class Range1(NamedUnaryOp):
+    """FPy node: range(stop)"""
+    __slots__ = ()
+
+class Range2(NamedBinaryOp):
+    """FPy node: range(start, stop)"""
+    __slots__ = ()
+
+class Range3(NamedTernaryOp):
+    """FPy node: range(start, stop, step)"""
     __slots__ = ()
 
 class Empty(NamedUnaryOp):
@@ -1682,12 +1716,12 @@ class Argument(Ast):
     """FPy AST: function argument"""
     __slots__ = ('name', 'type')
     name: Id
-    type: TypeAnn | None
+    type: TypeAnn
 
     def __init__(
         self,
         name: Id,
-        type: TypeAnn | None,
+        type: TypeAnn,
         loc: Location | None
     ):
         super().__init__(loc)
@@ -1695,7 +1729,11 @@ class Argument(Ast):
         self.type = type
 
     def is_equiv(self, other) -> bool:
-        return isinstance(other, Argument) and self.name == other.name
+        return (
+            isinstance(other, Argument)
+            and self.name == other.name
+            and self.type.is_equiv(other.type)
+        )
 
 class FuncDef(Ast):
     """FPy AST: function definition"""
