@@ -10,7 +10,7 @@ from .number import Context
 
 # avoids circular dependency issues (useful for type checking)
 if TYPE_CHECKING:
-    from .analysis.types import FunctionType
+    from .types import FunctionType
     from .interpret import Interpreter
 
 P = ParamSpec('P')
@@ -23,6 +23,13 @@ class Function(Generic[P, R]):
 
     This object is created by the `@fpy` decorator and represents
     a function in the FPy runtime.
+
+    Example::
+
+      @fp.fpy
+      def my_function(x: fp.Real) -> fp.Real:
+          return x * 2
+
     """
 
     ast: fpyast.FuncDef
