@@ -103,6 +103,7 @@ _BINARY_TABLE: dict[type[BinaryOp], Callable[..., Any]] = {
     Div: ops.div,
     Copysign: ops.copysign,
     Fdim: ops.fdim,
+    Mod: ops.mod,
     Fmod: ops.fmod,
     Remainder: ops.remainder,
     Hypot: ops.hypot,
@@ -149,21 +150,36 @@ class _Interpreter(Visitor):
             case _:
                 raise TypeError(f'Expected `Context` or `FPCoreContext`, got {ctx}')
 
-    def _arg_to_value(self, arg: Any):
+    def _is_value(self, x):
+        match x:
+            case bool() | Float() | Context():
+                return True
+            case tuple() | list():
+                return all(self._is_value(v) for v in x)
+            case _:
+                return False
+
+    def _cvt_arg(self, arg):
         match arg:
-            case Float():
+            case bool() | Float() | Context():
                 return arg
             case int():
                 return Float.from_int(arg, ctx=INTEGER, checked=False)
             case float():
                 return Float.from_float(arg, ctx=FP64, checked=False)
             case tuple():
-                return tuple(self._arg_to_value(x) for x in arg)
+                return tuple(self._cvt_arg(x) for x in arg)
             case list():
-                return [self._arg_to_value(x) for x in arg]
+                return [self._cvt_arg(x) for x in arg]
             case _:
                 return arg
 
+    def _arg_to_value(self, arg: Any):
+        if self._is_value(arg):
+            return arg
+        else:
+            return self._cvt_arg(arg)
+
     def _lookup(self, name: NamedId):
         try:
             return self.env[name]
@@ -244,13 +260,40 @@ class _Interpreter(Visitor):
             raise TypeError(f'expected a list, got {arr}')
         return Float.from_int(len(arr), ctx=INTEGER, checked=False)
 
-    def _apply_range(self, arg: Expr, ctx: Context):
-        stop = self._visit_expr(arg, ctx)
-        if not isinstance(stop, RealValue):
-            raise TypeError(f'expected a real number argument, got {stop}')
-        if not self._is_integer(stop):
-            raise ValueError(f'expected an integer argument, got {stop}')
-        return [Float.from_int(i, ctx=INTEGER, checked=False) for i in range(int(stop))]
+    def _apply_range(self, start: Expr | None, stop: Expr, step: Expr | None, ctx: Context):
+        if start is not None:
+            start_val = self._visit_expr(start, ctx)
+            if not isinstance(start_val, RealValue):
+                raise TypeError(f'expected a real number argument, got {start_val}')
+            if not self._is_integer(start_val):
+                raise ValueError(f'expected an integer argument, got {start_val}')
+            start_idx = int(start_val)
+        else:
+            start_idx = 0
+
+        stop_val = self._visit_expr(stop, ctx)
+        if not isinstance(stop_val, RealValue):
+            raise TypeError(f'expected a real number argument, got {stop_val}')
+        if not self._is_integer(stop_val):
+            raise ValueError(f'expected an integer argument, got {stop_val}')
+        stop_idx = int(stop_val)
+
+        if step is not None:
+            step_val = self._visit_expr(step, ctx)
+            if not isinstance(step_val, RealValue):
+                raise TypeError(f'expected a real number argument, got {step_val}')
+            if not self._is_integer(step_val):
+                raise ValueError(f'expected an integer argument, got {step_val}')
+            step_idx = int(step_val)
+            if step_idx == 0:
+                raise ValueError('step argument cannot be zero')
+        else:
+            step_idx = 1
+
+        return [
+            Float.from_int(i, ctx=INTEGER, checked=False)
+            for i in range(start_idx, stop_idx, step_idx)
+        ]
 
     def _apply_empty(self, arg: Expr, ctx: Context):
         size = self._visit_expr(arg, ctx)
@@ -379,8 +422,8 @@ class _Interpreter(Visitor):
                     return self._apply_not(e.arg, ctx)
                 case Len():
                     return self._apply_len(e.arg, ctx)
-                case Range():
-                    return self._apply_range(e.arg, ctx)
+                case Range1():
+                    return self._apply_range(None, e.arg, None, ctx)
                 case Empty():
                     return self._apply_empty(e.arg, ctx)
                 case Dim():
@@ -406,6 +449,8 @@ class _Interpreter(Visitor):
             match e:
                 case Size():
                     return self._apply_size(e.first, e.second, ctx)
+                case Range2():
+                    return self._apply_range(e.first, e.second, None, ctx)
                 case _:
                     raise RuntimeError('unknown operator', e)
 
@@ -423,7 +468,11 @@ class _Interpreter(Visitor):
                 raise TypeError(f'expected a real number argument, got {third}')
             return fn(first, second, third, ctx=ctx)
         else:
-            raise RuntimeError('unknown operator', e)
+            match e:
+                case Range3():
+                    return self._apply_range(e.first, e.second, e.third, ctx)
+                case _:
+                    raise RuntimeError('unknown operator', e)
 
     def _visit_naryop(self, e: NaryOp, ctx: Context):
         match e:
@@ -445,14 +494,20 @@ class _Interpreter(Visitor):
             # convert to int
             val = self._cvt_float(arg)
             if not val.is_integer():
-                raise TypeError(f'expected an integer argument for `{name}={arg}`')
+                raise ValueError(f'expected an integer argument for `{name}={arg}`')
             return int(val)
         elif ty is float:
             # convert to float
-            raise NotImplementedError(arg, ty)
+            val = self._cvt_float(arg)
+            if not FP64.representable_under(val):
+                raise ValueError(f'argument for `{name}={arg}` is not representable as a float')
+            return float(val)
         elif ty is RealFloat:
             # convert to RealFloat
-            raise NotImplementedError(arg, ty)
+            val = self._cvt_float(arg)
+            if val.is_nar():
+                raise ValueError(f'argument for `{name}={arg}` cannot be Inf/NaN')
+            return val.as_real()
         else:
             # don't apply a conversion
             return arg
@@ -622,6 +677,18 @@ class _Interpreter(Visitor):
         array[slices[-1]] = val
         return array
 
+    def _bind(self, target: Id | TupleBinding, val: Value) -> None:
+        match target:
+            case NamedId():
+                self.env[target] = val
+            case TupleBinding():
+                if not isinstance(val, tuple):
+                    raise TypeError(f'can only unpack tuples, got `{val}` for `{target}`')
+                if len(target.elts) != len(val):
+                    raise NotImplementedError(f'unpacking {len(val)} values into {len(target.elts)}')
+                for elt, v in zip(target.elts, val):
+                    self._bind(elt, v)
+
     def _apply_comp(
         self,
         bindings: list[tuple[Id | TupleBinding, Expr]],
@@ -637,13 +704,7 @@ class _Interpreter(Visitor):
             if not isinstance(array, list):
                 raise TypeError(f'expected a list, got {array}')
             for val in array:
-                match target:
-                    case NamedId():
-                        self.env[target] = val
-                    case TupleBinding():
-                        if not isinstance(val, tuple):
-                            raise TypeError(f'can only unpack tuples, got `{val}` for `{target}`')
-                        self._unpack_tuple(target, val, ctx)
+                self._bind(target, val)
                 self._apply_comp(bindings[1:], elt, ctx, elts)
 
     def _visit_list_comp(self, e: ListComp, ctx: Context):
@@ -659,27 +720,9 @@ class _Interpreter(Visitor):
             raise TypeError(f'expected a boolean, got {cond}')
         return self._visit_expr(e.ift if cond else e.iff, ctx)
 
-    def _unpack_tuple(self, binding: TupleBinding, val: tuple, ctx: Context) -> None:
-        if len(binding.elts) != len(val):
-            raise NotImplementedError(f'unpacking {len(val)} values into {len(binding.elts)}')
-        for elt, v in zip(binding.elts, val):
-            match elt:
-                case NamedId():
-                    self.env[elt] = v
-                case TupleBinding():
-                    if not isinstance(val, tuple):
-                        raise TypeError(f'can only unpack tuples, got `{val}` for `{binding}`')
-                    self._unpack_tuple(elt, v, ctx)
-
     def _visit_assign(self, stmt: Assign, ctx: Context) -> None:
         val = self._visit_expr(stmt.expr, ctx)
-        match stmt.target:
-            case NamedId():
-                self.env[stmt.target] = val
-            case TupleBinding():
-                if not isinstance(val, tuple):
-                    raise TypeError(f'can only unpack tuples, got `{val}` for `{stmt.target}`')
-                self._unpack_tuple(stmt.target, val, ctx)
+        self._bind(stmt.target, val)
 
     def _visit_indexed_assign(self, stmt: IndexedAssign, ctx: Context) -> None:
         # lookup the array
@@ -745,14 +788,8 @@ class _Interpreter(Visitor):
             raise TypeError(f'expected a list, got {iterable}')
         # iterate over each element
         for val in iterable:
-            match stmt.target:
-                case NamedId():
-                    self.env[stmt.target] = val
-                case TupleBinding():
-                    if not isinstance(val, tuple):
-                        raise TypeError(f'can only unpack tuples, got `{val}` for `{stmt.target}`')
-                    self._unpack_tuple(stmt.target, val, ctx)
             # evaluate the body of the loop
+            self._bind(stmt.target, val)
             self._visit_block(stmt.body, ctx)
 
     def _visit_attribute(self, e: Attribute, ctx: Context):
@@ -822,33 +859,38 @@ class _Interpreter(Visitor):
     def eval(self, func: FuncDef, args: Collection[Any], ctx: Context):
         # check arity
         if len(args) != len(func.args):
-            raise TypeError(f'Expected {len(func.args)} arguments, got {len(args)}')
+            raise RuntimeError(f'{func.name}: expected {len(func.args)} arguments, got {len(args)}')
 
         # possibly override the context
         eval_ctx = self._eval_ctx(ctx)
 
         # process arguments and add to environment
         for val, arg in zip(args, func.args):
+            # convert the argument
+            x = self._arg_to_value(val)
+
+            # check the argument type
             match arg.type:
-                case AnyTypeAnn() | None:
-                    x = self._arg_to_value(val)
-                    if isinstance(arg.name, NamedId):
-                        self.env[arg.name] = x
+                case AnyTypeAnn():
+                    pass
+                case BoolTypeAnn():
+                    if not isinstance(x, bool):
+                        raise TypeError(f'argument expects a boolean, got data `{x}`')
                 case RealTypeAnn():
-                    x = self._arg_to_value(val)
                     if not isinstance(x, RealValue):
                         raise TypeError(f'argument expects a real number, got data `{val}`')
-                    if isinstance(arg.name, NamedId):
-                        self.env[arg.name] = x
-                case TensorTypeAnn():
+                case TupleTypeAnn():
+                    if not isinstance(x, tuple):
+                        raise NotImplementedError(f'argument is a tuple, got data {val}')
+                case ListTypeAnn() | SizedTensorTypeAnn():
                     # TODO: check shape
-                    x = self._arg_to_value(val)
                     if not isinstance(x, list):
                         raise NotImplementedError(f'argument is a list, got data {val}')
-                    if isinstance(arg.name, NamedId):
-                        self.env[arg.name] = x
                 case _:
-                    raise NotImplementedError(f'unknown argument type {arg.type}')
+                    pass
+
+            if isinstance(arg.name, NamedId):
+                self.env[arg.name] = x
 
         # evaluate the function body
         return self._visit_function(func, eval_ctx)
