@@ -377,6 +377,11 @@ class EFloatContext(EncodableContext):
             raise TypeError(f'Expected a representable \'Float\', got \'{type(x)}\' for x={x}')
         return self._mpb_ctx.to_ordinal(x, infval=infval)
 
+    def to_fractional_ordinal(self, x: Float):
+        if not isinstance(x, Float):
+            raise TypeError(f'Expected \'Float\', got \'{type(x)}\' for x={x}')
+        return self._mpb_ctx.to_fractional_ordinal(x)
+
     def from_ordinal(self, x: int, infval: bool = False) -> Float:
         y = self._mpb_ctx.from_ordinal(x, infval=infval)
         y._ctx = self
@@ -783,10 +788,18 @@ def _ext_to_mpb(
                     # NaN is in the last binade, and the maximum value is
                     # the usual IEEE 754 maximum value
                     maxval = RealFloat(c=bitmask(p), exp=expmax)
-            elif p == 2 and enable_inf:
-                # NaN and Inf occupy the last binade,
-                # so the maximum value is the usual IEEE 754 maximum value
-                maxval = RealFloat(c=bitmask(p), exp=expmax)
+            elif p == 2:
+                if enable_inf:
+                    # NaN and Inf occupy the last binade,
+                    # so the maximum value is the usual IEEE 754 maximum value
+                    maxval = RealFloat(c=bitmask(p), exp=expmax)
+                elif es == 0:
+                    # 0 and NaN occupy the last binade
+                    maxval = RealFloat.from_int(0)
+                else:
+                    # only NaN occupies the last binade
+                    c = bitmask(p) - 1
+                    maxval = RealFloat(c=c, exp=expmax+1)
             else:
                 # maximum value is in the last binade
                 if enable_inf:
@@ -820,7 +833,8 @@ def _ext_to_mpb(
 
     # for es == 0, the exponent may be too low
     # we should normalize and ensure the exponent is in range
-    maxval = maxval.normalize(p, expmin - 1)
+    if es == 0:
+        maxval = maxval.normalize(p, expmin - 1)
 
     # create the related MPB context
     return MPBFloatContext(p, emin, maxval, rm, overflow, num_randbits)
