@@ -9,7 +9,6 @@ from fractions import Fraction
 
 from ..utils import DEFAULT, DefaultOr, bitmask, default_repr
 from .context import EncodableContext, Context
-from .gmp import mpfr_value
 from .mp_float import MPFloatContext
 from .round import OverflowMode, RoundingMode, RoundingDirection
 from .number import Float, RealFloat
@@ -120,6 +119,9 @@ class ExpContext(EncodableContext):
         *,
         inf_value: Float | None = None
     ):
+        if overflow == OverflowMode.WRAP:
+            raise ValueError('OverflowMode.WRAP is not supported for ExpContext')
+
         emin, emax, mp_ctx, inf_value = _compute_params(nbits, eoffset, rm, overflow, inf_value)
         self.nbits = nbits
         self.eoffset = eoffset
@@ -184,6 +186,10 @@ class ExpContext(EncodableContext):
             inf_value = self.inf_value
         if kwargs:
             raise TypeError(f'Unexpected parameters {kwargs} for ExpContext')
+
+        if overflow == OverflowMode.WRAP:
+            raise ValueError('OverflowMode.WRAP is not supported for ExpContext')
+
         return ExpContext(
             nbits,
             eoffset,
@@ -386,6 +392,11 @@ class ExpContext(EncodableContext):
         x = self._round_prepare(x)
         return self._round_at(x, n, exact)
 
+    def _to_ordinal(self, x: RealFloat | Float) -> int:
+        if isinstance(x, Float) and x.isnan:
+            raise ValueError(f'can only convert a finite value x={x}')
+        return x.e + self.ebias
+
     def to_ordinal(self, x: Float, infval: bool = False) -> int:
         if not isinstance(x, Float):
             raise TypeError(f'Expected a representable \'Float\', got \'{type(x)}\' for x={x}')
@@ -396,7 +407,42 @@ class ExpContext(EncodableContext):
         if x.isnan:
             # NaN
             raise ValueError(f'can only convert a finite value x={x}')
-        return self.encode(x)
+        return self._to_ordinal(x)
+
+    def to_fractional_ordinal(self, x: Float):
+        if not isinstance(x, Float):
+            raise TypeError(f'Expected \'Float\', got \'{type(x)}\' for x={x}')
+        if x.is_nar():
+            # NaN or Inf
+            raise ValueError(f'Expected a finite value for x={x}')
+
+        if self.representable_under(x):
+            # representable value
+            return Fraction(self.to_ordinal(x))
+        else:
+            # not representable value
+            # step 1. compute the nearest values, above and below `x`
+            xr = x.as_real()
+            above = xr.round(self.pmax, rm=RoundingMode.RTP)
+            below = xr.round(self.pmax, rm=RoundingMode.RTN)
+            if above == below:
+                # not representable, but rounds to the same value without bounds
+                return Fraction(self._to_ordinal(above))
+
+            # step 2. ordinal space is linear between two adjacent values;
+            # compute the linear interpolation factor
+            # TODO: is this true? one could argue the space is exponential
+            delta_x: RealFloat = xr - below
+            delta: RealFloat = above - below
+            t = delta_x.as_rational() / delta.as_rational()
+
+            # step 3. map one endpoint to the ordinals (they are one apart)
+            # TODO: what if `below` is not encodable?
+            below_ord = self._to_ordinal(below)
+
+            # step 4. apply linear interpolation
+            return Fraction(below_ord) + t
+
 
     def from_ordinal(self, x: int, infval: bool = False) -> Float:
         if not isinstance(x, int):
