@@ -4,7 +4,7 @@ Global variables for the `fpy2.number` module.
 
 from . import number
 
-from typing import Callable, Optional, TypeAlias, Union
+from typing import Callable, TypeAlias, Union
 
 # avoids circular dependency issues (useful for type checking)
 Float: TypeAlias = 'number.Float'
@@ -14,8 +14,8 @@ RealFloat: TypeAlias = 'number.RealFloat'
 _FloatCvt: TypeAlias = Callable[[Union[Float, RealFloat]], float]
 _StrCvt: TypeAlias = Callable[[Union[Float, RealFloat]], str]
 
-_current_float_converter: Optional[_FloatCvt] = None
-_current_str_converter: Optional[_StrCvt] = None
+_current_float_converter: _FloatCvt | None = None
+_current_str_converter: _StrCvt | None = None
 
 
 def get_current_float_converter() -> _FloatCvt:
