@@ -3,14 +3,11 @@ This module defines floating-point numbers as implemented by MPFR,
 that is, multi-precision floating-point numbers. Hence, "MP."
 """
 
-from fractions import Fraction
-
 from ..utils import bitmask, default_repr, DefaultOr, DEFAULT
 
 from .context import Context
 from .number import RealFloat, Float
 from .round import RoundingMode
-from .gmp import mpfr_value
 
 @default_repr
 class MPFloatContext(Context):
@@ -102,13 +99,15 @@ class MPFloatContext(Context):
             case _:
                 raise TypeError(f'Expected \'RealFloat\' or \'Float\', got \'{type(x)}\' for x={x}')
 
+        # value is finite and non-zero
         # precision is possibly out of bounds
         # check if the value can be normalized with fewer digits
-        p_over = x.p - self.pmax
-        if p_over < 0:
+        if x.p <= self.pmax:
             # precision is within bounds
             return True
         else:
+            # precision is out of bounds, so check if the excess bits are all zero
+            p_over = x.p - self.pmax
             c_lost = x.c & bitmask(p_over) # bits that would be lost via normalization
             return c_lost == 0
 
