@@ -3,15 +3,12 @@ This module defines fixed-pont numbers with a maximum value,
 that is, multiprecision and bounded. Hence "MP-B".
 """
 
-from fractions import Fraction
-
 from ..utils import default_repr, DefaultOr, DEFAULT
 
 from .context import Context, SizedContext
 from .mp_fixed import MPFixedContext
 from .number import RealFloat, Float
 from .round import RoundingMode, OverflowMode
-from .gmp import mpfr_value
 
 
 @default_repr
@@ -329,14 +326,14 @@ class MPBFixedContext(SizedContext):
             case Float():
                 if x.isnan:
                     if self.enable_nan:
-                        return Float(isnan=True, ctx=self)
+                        return Float(s=x.s, isnan=True, ctx=self)
                     elif self.nan_value is None:
                         raise ValueError('Cannot round NaN under this context')
                     else:
                         return Float(x=self.nan_value, ctx=self)
                 elif x.isinf:
                     if self.enable_inf:
-                        return Float(isinf=True, ctx=self)
+                        return Float(s=x.s, isinf=True, ctx=self)
                     elif self.inf_value is None:
                         raise ValueError('Cannot round infinity under this context')
                     else:
@@ -412,6 +409,11 @@ class MPBFixedContext(SizedContext):
             # finite, real
             return self._mp_ctx.to_ordinal(x)
 
+    def to_fractional_ordinal(self, x: Float):
+        if not isinstance(x, Float):
+            raise TypeError(f'Expected \'Float\', got \'{type(x)}\' for x={x}')
+        return self._mp_ctx.to_fractional_ordinal(x)
+
     def from_ordinal(self, x: int, infval: bool = False) -> Float:
         if not isinstance(x, int):
             raise TypeError(f'Expected \'int\' for x={x}, got {type(x)}')
@@ -440,12 +442,16 @@ class MPBFixedContext(SizedContext):
     def minval(self, s: bool = False) -> Float:
         if not isinstance(s, bool):
             raise TypeError(f'Expected \'bool\' for s={s}, got {type(s)}')
+        if s and not self.neg_maxval.is_negative():
+            raise ValueError('negative values are not representable')
         return Float(x=self._mp_ctx.minval(s=s), ctx=self)
 
     def maxval(self, s: bool = False) -> Float:
         if not isinstance(s, bool):
             raise TypeError(f'Expected \'bool\' for s={s}, got {type(s)}')
         if s:
+            if not self.neg_maxval.is_negative():
+                raise ValueError('negative values are not representable')
             return Float(x=self.neg_maxval, ctx=self)
         else:
             return Float(x=self.pos_maxval, ctx=self)
