@@ -4,15 +4,12 @@ but with subnormalization and a maximum value, that is multi-precision
 and bounded. Hence, "MP-B."
 """
 
-from fractions import Fraction
-
 from ..utils import default_repr, DefaultOr, DEFAULT
 
 from .context import Context, SizedContext
 from .number import RealFloat, Float
 from .mps_float import MPSFloatContext
 from .round import RoundingMode, RoundingDirection, OverflowMode
-from .gmp import mpfr_value
 
 
 @default_repr
@@ -359,6 +356,10 @@ class MPBFloatContext(SizedContext):
             # finite, real
             return self._mps_ctx.to_ordinal(x)
 
+    def to_fractional_ordinal(self, x: Float):
+        if not isinstance(x, Float):
+            raise TypeError(f'Expected \'Float\', got \'{type(x)}\' for x={x}')
+        return self._mps_ctx.to_fractional_ordinal(x)
 
     def from_ordinal(self, x, infval = False):
         if not isinstance(x, int):
