@@ -211,14 +211,40 @@ class RealFloat(numbers.Rational):
         ord = self.compare(other)
         return ord == Ordering.GREATER or ord == Ordering.EQUAL
 
-    def __add__(self, other: 'RealFloat'):
+    def __neg__(self):
+        """
+        Unary minus.
+
+        Returns this `RealFloat` with opposite sign (`self.s`)
+        even when `self.is_zero()`.
+        """
+        return RealFloat(s=not self._s, x=self)
+
+    def __pos__(self):
+        """
+        Unary plus. 
+
+        Returns a copy of `self`.
+        """
+        return RealFloat(x=self)
+
+    def __add__(self, other):
         """
         Adds `self` and `other` exactly.
 
         This operation never fails when `other` is a `RealFloat`.
         """
-        if not isinstance(other, RealFloat):
-            raise TypeError(f'unsupported operand type(s) for +: \'RealFloat\' and \'{type(other)}\'')
+        match other:
+            case RealFloat():
+                pass
+            case int():
+                other = RealFloat.from_int(other)
+            case float():
+                other = RealFloat.from_float(other)
+            case Fraction():
+                other = RealFloat.from_rational(other)
+            case _:
+                raise TypeError(f'unsupported operand type(s) for +: \'RealFloat\' and \'{type(other)}\'')
 
         if self._c == 0:
             # 0 + b = b
@@ -250,35 +276,32 @@ class RealFloat(numbers.Rational):
             # return the result
             return RealFloat(s=s, exp=exp, c=c)
 
-
     def __radd__(self, other):
         return self + other
 
-    def __neg__(self):
-        """
-        Unary minus.
+    def __sub__(self, other):
+        return self + (-other)
+    
+    def __rsub__(self, other):
+        return (-self) + other
 
-        Returns this `RealFloat` with opposite sign (`self.s`)
-        even when `self.is_zero()`.
-        """
-        return RealFloat(s=not self._s, x=self)
-
-    def __pos__(self):
-        """
-        Unary plus. 
-
-        Returns a copy of `self`.
-        """
-        return RealFloat(x=self)
-
-    def __mul__(self, other: 'RealFloat'):
+    def __mul__(self, other):
         """
         Multiplies `self` and `other` exactly.
 
         This operation never fails when `other` is a `RealFloat`.
         """
-        if not isinstance(other, RealFloat):
-            raise TypeError(f'unsupported operand type(s) for *: \'RealFloat\' and \'{type(other)}\'')
+        match other:
+            case RealFloat():
+                pass
+            case int():
+                other = RealFloat.from_int(other)
+            case float():
+                other = RealFloat.from_float(other)
+            case Fraction():
+                other = RealFloat.from_rational(other)
+            case _:
+                raise TypeError(f'unsupported operand type(s) for +: \'RealFloat\' and \'{type(other)}\'')
 
         s = self._s != other._s
         if self._c == 0 or other._c == 0:
@@ -322,7 +345,7 @@ class RealFloat(numbers.Rational):
             return RealFloat(s=s, exp=exp, c=c)
 
     def __rpow__(self, base):
-        raise NotImplementedError
+        raise TypeError(f'unsupported operand type(s) for **: \'{type(base)}\' and \'RealFloat\'')
 
     def __abs__(self):
         """
@@ -333,16 +356,16 @@ class RealFloat(numbers.Rational):
         return RealFloat(s=False, x=self)
 
     def __trunc__(self) -> int:
-        raise NotImplementedError('do not call directly')
+        return int(self.round(min_n=-1, rm=RoundingMode.RTZ))
 
     def __floor__(self) -> int:
-        raise NotImplementedError('do not call directly')
+        return int(self.round(min_n=-1, rm=RoundingMode.RTN))
 
     def __ceil__(self) -> int:
-        raise NotImplementedError('do not call directly')
+        return int(self.round(min_n=-1, rm=RoundingMode.RTP))
 
     def __round__(self, ndigits=None) -> int:
-        raise NotImplementedError('do not call directly')
+        return int(self.round(min_n=-1, rm=RoundingMode.RNE))
 
     def __floordiv__(self, other):
         raise NotImplementedError('division cannot be implemented exactly')
@@ -606,6 +629,9 @@ class RealFloat(numbers.Rational):
             _, lo = self.split(n)
             return lo.is_zero()
         """
+        if not isinstance(n, int):
+            raise TypeError(f'expected \'int\' for n, got {n}')
+
         if self.is_zero():
             return True
 
@@ -634,52 +660,73 @@ class RealFloat(numbers.Rational):
         if not isinstance(n, int):
             raise ValueError('expected an integer', n)
 
-        # special case: 0 has no digits set
-        if self.is_zero():
+        # compute digit offset from `self.exp`
+        offset = n - self._exp
+
+        # outside the region of significance
+        if offset < 0 or offset >= self.p:
             return False
 
-        # below the region of significance
-        if n < self._exp:
-            return False
+        # test the `offset`-th bit of `self.c`
+        return (self._c & (1 << offset)) != 0
 
-        # above the region of significane
-        if n > self.e:
-            return False
 
-        idx = n - self._exp
-        bit = self._c & (1 << idx)
-        return bit != 0
-
-    def normalize(self, p: int, n: int | None = None):
+    def normalize(self, p: int | None = None, n: int | None = None):
         """
-        Returns a copy of `self` that has exactly `p` bits of precision.
-        Optionally, specify `n` to ensure that if `y = x.normalize(p, n)`,
-        then `y.exp > n` or `y` is zero.
+        Returns a value numerically equivalent to `self` based on
+        precision `p` and position `n`:
 
-        For non-zero values, raises a `ValueError` if any significand digits
-        are shifted off, i.e., `x != x.normalize(p, n)`.
+        - `None, None`: a copy of `self`, i.e., `self.exp == self.normalize().exp`, etc.
+        - `p, None`: a copy of `self` that has exactly `p` bits of precision.
+        - `None, n`: a copy of `self` where `self.exp == n + 1`.
+        - `p, n`: a copy of `self` such that `self.exp >= n + 1` and
+            has maximal precision up to `p` bits.
+
+        Raises a `ValueError` if no such value exists, i.e.,
+        if the value cannot be represented with the given `p` and `n`.
         """
-        if not isinstance(p, int) or p < 0:
-            raise ValueError('expected a non-negative integer', p)
 
-        # special case: 0 has no precision
-        if self.is_zero():
-            return RealFloat()
-
-        # compute maximum shift and resulting exponent
-        shift = p - self.p
-        exp = self._exp - shift
-
-        # test if exponent is below `n`
-        if n is not None and exp <= n:
-            # too small, so adjust accordingly
-            expmin = n + 1
-            adjust = expmin - exp
-            shift -= adjust
-            exp += adjust
+        match p, n:
+            case None, None:
+                # return a copy of self
+                return RealFloat(self._s, self._exp, self._c)
+            case int(), None:
+                # normalize to precision p
+                if p < 0:
+                    raise ValueError(f'precision must be non-negative: p={p}')
+                # compute maximum shift and resulting exponent
+                shift = p - self.p
+                exp = self._exp - shift
+            case None, int():
+                # normalize to absolute position `n`
+                exp = n + 1
+                shift = self._exp - exp
+            case int(), int():
+                # normalize to precision p and position n
+                # prefer absolute precision constraint over position constraint
+                if p < 0:
+                    raise ValueError(f'precision must be non-negative: p={p}')
+                # compute maximum shift and resulting exponent
+                shift = p - self.p
+                exp = self._exp - shift
+                # check if exponent is too small, adjust accordingly
+                if exp <= n:
+                    expmin = n + 1
+                    adjust = expmin - exp
+                    shift -= adjust
+                    exp += adjust
+            case _:
+                if p is not None and not isinstance(p, int):
+                    raise TypeError(f'expected \'int\' or \'None\' for p, got {type(p)}')
+                if n is not None and not isinstance(n, int):
+                    raise TypeError(f'expected \'int\' or \'None\' for n, got {type(n)}')
+                raise RuntimeError('unreachable')
 
         # compute new significand `c`
-        if shift >= 0:
+        if shift == 0:
+            # no shifting
+            c = self._c
+        elif shift > 0:
             # shifting left by a non-negative amount
             c = self._c << shift
         else:
@@ -690,7 +737,6 @@ class RealFloat(numbers.Rational):
             if (self._c & bitmask(shift)) != 0:
                 raise ValueError(f'shifting off digits: p={p}, n={n}, x={self}')
 
-        # return result
         return RealFloat(self._s, exp, c)
 
 
@@ -1382,6 +1428,213 @@ class Float:
         ord = self.compare(other)
         return ord == Ordering.GREATER or ord == Ordering.EQUAL
 
+    def __neg__(self):
+        """
+        Unary minus.
+
+        Returns this `Float` with opposite sign (`self.s`)
+        and no context (`self.ctx is None`).
+        """
+        return Float(s=not self._real._s, x=self, ctx=None)
+
+    def __pos__(self):
+        """
+        Unary plus.
+
+        Returns this `Float` with no context (`self.ctx is None`).
+        """
+        return Float(s=False, x=self, ctx=None)
+
+    def __abs__(self):
+        """
+        Absolute value.
+
+        Returns this `Float` without the sign (`self.s = False`)
+        with no context (`self.ctx is None`).
+        """
+        return Float(s=False, x=self, ctx=None)
+
+    def __add__(self, other):
+        """
+        Addition: `self + other`.
+
+        Returns the exact sum of `self` and `other` as a new `Float`.
+        The result has no context (`self.ctx is None`).
+        """
+        match other:
+            case Float():
+                pass
+            case RealFloat():
+                other = Float.from_real(other)
+            case int():
+                other = Float.from_int(other)
+            case float():
+                other = Float.from_float(other)
+            case Fraction():
+                other = Float.from_rational(other)
+            case _:
+                raise TypeError(f'unsupported operand type(s) for +: \'RealFloat\' and \'{type(other)}\'')
+
+        if self.isnan or other.isnan:
+            # either is NaN
+            return Float(isnan=True)
+        elif self.isinf:
+            # self is Inf
+            if other.isinf:
+                # other is Inf
+                if self.s == other.s:
+                    # Inf + Inf
+                    return Float(s=self.s, isinf=True)
+                else:
+                    # Inf - Inf
+                    return Float(isnan=True)
+            else:
+                # other is finite, Inf + y = Inf
+                return Float(s=self.s, isinf=True)
+        elif other.isinf:
+            # self is finite, x + Inf = Inf
+            return Float(s=other.s, isinf=True)
+        else:
+            # both are finite
+            return Float.from_real(self._real + other._real)
+
+    def __radd__(self, other):
+        return self + other
+
+    def __sub__(self, other):
+        return self + (-other)
+    
+    def __rsub__(self, other):
+        return (-self) + other
+
+    def __mul__(self, other):
+        """
+        Multiplication: `self * other`.
+
+        Returns the exact product of `self` and `other` as a new `Float`.
+        The result has no context (`self.ctx is None`).
+        """
+        match other:
+            case Float():
+                pass
+            case RealFloat():
+                other = Float.from_real(other)
+            case int():
+                other = Float.from_int(other)
+            case float():
+                other = Float.from_float(other)
+            case Fraction():
+                other = Float.from_rational(other)
+            case _:
+                raise TypeError(f'unsupported operand type(s) for +: \'RealFloat\' and \'{type(other)}\'')
+
+        if self.isnan or other.isnan:
+            # either is NaN
+            return Float(isnan=True)
+        elif self.isinf:
+            # self is Inf
+            if other.is_zero():
+                # Inf * 0 = NaN
+                return Float(isnan=True)
+            else:
+                # Inf * y = Inf
+                s = self._real._s != other._real._s
+                return Float(s=s, isinf=True)
+        elif other.isinf:
+            # y is Inf
+            if self.is_zero():
+                # 0 * Inf = NaN
+                return Float(isnan=True)
+            else:
+                # x * Inf = Inf
+                s = self._real._s != other._real._s
+                return Float(s=s, isinf=True)
+        else:
+            # both are finite
+            return Float.from_real(self._real * other._real)
+
+    def __rmul__(self, other):
+        return self * other
+
+    def __truediv__(self, other: 'Real'):
+        if TYPE_CHECKING:
+            return Float()
+        else:
+            raise RuntimeError('FPy runtime: do not call directly')
+
+    def __rtruediv__(self, other: 'Real'):
+        if TYPE_CHECKING:
+            return Float()
+        else:
+            raise RuntimeError('FPy runtime: do not call directly')
+
+    def __pow__(self, exponent: 'Real'):
+        """
+        Raising `self` by `exponent` exactly.
+
+        This operation is only valid for `exponent` of type `int` with `exponent >= 0`.
+        """
+        if not isinstance(exponent, int):
+            raise TypeError(f'unsupported operand type(s) for **: \'RealFloat\' and \'{type(exponent)}\'')
+        if exponent < 0:
+            raise ValueError('negative exponent unsupported; cannot be implemented exactly')
+
+        if self.is_nar():
+            s = self._real._s and (exponent % 2 != 0)
+            return Float(x=self, s=s, ctx=None)
+        else:
+            return Float.from_real(self._real ** exponent)
+
+    def __rpow__(self, base: 'Real'):
+        if TYPE_CHECKING:
+            return Float()
+        else:
+            raise RuntimeError('FPy runtime: do not call directly')
+
+    def __trunc__(self):
+        if self.is_nar():
+            raise ValueError('cannot round infinity or NaN')
+        return self._real.__trunc__()
+
+    def __floor__(self):
+        if self.is_nar():
+            raise ValueError('cannot round infinity or NaN')
+        return self._real.__floor__()
+
+    def __ceil__(self):
+        if self.is_nar():
+            raise ValueError('cannot round infinity or NaN')
+        return self._real.__ceil__()
+
+    def __round__(self, *args, **kwargs):
+        if self.is_nar():
+            raise ValueError('cannot round infinity or NaN')
+        return self._real.__round__()
+
+    def __floordiv__(self, other: 'Real'):
+        if TYPE_CHECKING:
+            return Float()
+        else:
+            raise RuntimeError('FPy runtime: do not call directly')
+
+    def __rfloordiv__(self, other: 'Real'):
+        if TYPE_CHECKING:
+            return Float()
+        else:
+            raise RuntimeError('FPy runtime: do not call directly')
+
+    def __mod__(self, other: 'Real'):
+        if TYPE_CHECKING:
+            return Float()
+        else:
+            raise RuntimeError('FPy runtime: do not call directly')
+
+    def __rmod__(self, other: 'Real'):
+        if TYPE_CHECKING:
+            return Float()
+        else:
+            raise RuntimeError('FPy runtime: do not call directly')
+
     def __float__(self):
         """
         Casts this value exactly to a native Python float.
@@ -1579,6 +1832,25 @@ class Float:
             raise ValueError(f'Float values without a context cannot be normalized: self={self}')
         return self._ctx.normal_under(self)
 
+    def is_more_significant(self, n: int) -> bool:
+        """
+        Returns `True` iff this value only has significant digits above `n`,
+        that is, every non-zero digit in the number is more significant than `n`.
+
+        Raises a `ValueError` when `self.is_nar()`.
+
+        This method is equivalent to::
+
+            assert not self.is_nar()
+            _, lo = self.split(n)
+            return lo.is_zero()
+        """
+        if not isinstance(n, int):
+            raise TypeError(f'expected \'int\' for n, got {n}')
+        if self.is_nar():
+            raise ValueError('cannot check significance of infinity or NaN')
+        return self._real.is_more_significant(n)
+
     def as_rational(self) -> Fraction:
         """
         Converts this value to a `Fraction` representing the same value.
@@ -1679,7 +1951,7 @@ class Float:
         else:
             xr = RealFloat.from_float(x)
             return Float.from_real(xr, ctx, checked)
-        
+
     @staticmethod
     def from_rational(x: Fraction, ctx: Context | None = None, checked: bool = True):
         """
@@ -1700,15 +1972,65 @@ class Float:
             raise ValueError('cannot convert infinity or NaN to real')
         return self._real
 
-    def normalize(self):
+    def normalize(self, p: int | None = None, n: int | None = None):
         """
-        Returns the canonical reprsentation of this number.
+        Returns a value numerically equivalent to `self` based on
+        precision `p` and position `n`:
 
-        Raises a `ValueError` when `self.ctx is None`.
+        - `None, None`: the canonical representation of `self` under `self.ctx`.
+        - `p, None`: a copy of `self` that has exactly `p` bits of precision.
+        - `None, n`: a copy of `self` where `self.exp == n + 1`.
+        - `p, n`: a copy of `self` such that `self.exp >= n + 1` and
+            has maximal precision up to `p` bits.
+
+        Raises a `ValueError` if no such value exists, i.e.,
+        if the value cannot be represented with the given `p` and `n`.
         """
-        if self._ctx is None:
-            raise ValueError(f'cannot normalize without a context: self={self}')
-        return self._ctx.normalize(self)
+        if p is not None:
+            if not isinstance(p, int):
+                raise TypeError(f'expected \'int\' for p={p}, got {type(p)}')
+            if p < 0:
+                raise ValueError(f'expected non-negative integer for p={p}')
+        if n is not None and not isinstance(n, int):
+            raise TypeError(f'expected \'int\' for n={n}, got {type(n)}')
+
+        if p is None and n is None:
+            # normalize under the context
+            if self._ctx is None:
+                raise ValueError(f'cannot normalize without a context: self={self}')
+            return self._ctx.normalize(self)
+        else:
+            # normalize with given parameters
+            if self.isnan:
+                return Float(isnan=True, s=self.s, ctx=self._ctx)
+            elif self.isinf:
+                return Float(isinf=True, s=self.s, ctx=self._ctx)
+            else:
+                xr = self._real.normalize(p, n)
+                return Float(x=xr, ctx=self._ctx)
+
+    def split(self, n: int):
+        """
+        Splits `self` into two `Float` values where the first value represents
+        the digits above `n` and the second value represents the digits below
+        and including `n`.
+        """
+        if not isinstance(n, int):
+            raise TypeError(f'expected \'int\' for n={n}, got {type(n)}')
+
+        if self.isnan:
+            hi = Float(isnan=True, s=self.s, ctx=self._ctx)
+            lo = Float(isnan=True, s=self.s, ctx=self._ctx)
+        elif self.isinf:
+            hi = Float(isinf=True, s=self.s, ctx=self._ctx)
+            lo = Float(isinf=True, s=self.s, ctx=self._ctx)
+        else:
+            hr, lr = self._real.split(n)
+            hi = Float(x=hr, ctx=self._ctx)
+            lo = Float(x=lr, ctx=self._ctx)
+
+        return hi, lo
+
 
     def round(self, ctx: Context):
         """
@@ -1788,137 +2110,6 @@ class Float:
                 case _:
                     return False
 
-    def __add__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __radd__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __sub__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __rsub__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __neg__(self):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __pos__(self):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __mul__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __rmul__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __truediv__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __rtruediv__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __pow__(self, exponent: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __rpow__(self, base: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __abs__(self):
-        # these guards are due to a limitation of mypy
-        # we cannot just add `Float` as the return type.
-        # In the case of `x : Float | int`, mypy will determine
-        # that `abs(x) : Join[Float, int] = object` which is incorrect.
-        # pyright does not join but unions instead
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __trunc__(self):
-        if TYPE_CHECKING:
-            return 0
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __floor__(self):
-        if TYPE_CHECKING:
-            return 0
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __ceil__(self):
-        if TYPE_CHECKING:
-            return 0
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __round__(self, *args, **kwargs):
-        if TYPE_CHECKING:
-            return 0
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __floordiv__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __rfloordiv__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __mod__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
-    def __rmod__(self, other: 'Real'):
-        if TYPE_CHECKING:
-            return Float()
-        else:
-            raise RuntimeError('FPy runtime: do not call directly')
-
 ###########################################################
 # Type Aliases
 
