@@ -25,10 +25,14 @@ from .float_params import (
 from .fractions import digits_to_fraction, decnum_to_fraction, hexnum_to_fraction, is_dyadic
 from .gensym import Gensym
 from .identifier import Id, NamedId, UnderscoreId, SourceId
-from .inspect import has_keyword
+from .inspect import getfunclines, has_keyword
 from .iterator import sliding_window
+from .loader import get_module_source, install_caching_loader
 from .location import Location
 from .ordering import Ordering
 from .string import pythonize_id
 from .uninit import UNINIT
 from .unionfind import Unionfind
+
+# install custom loader
+install_caching_loader()
