@@ -29,3 +29,6 @@ class Location:
         if not isinstance(other, Location):
             return False
         return self.__key() == other.__key()
+
+    def format(self):
+        return f'`{self.source}:{self.start_line}:{self.start_column}'
